@@ -10,10 +10,8 @@
 #include <string.h>
 #include <unistd.h>
 
-#include <lz4.h>
-#include <openssl/sha.h>
-
 #include "arg.h"
+#include "blake2.h"
 #include "dedup.h"
 
 #define SNAPSF ".snapshots"
@@ -41,55 +39,6 @@ static unsigned long long cache_misses;
 int verbose;
 char *argv0;
 
-static size_t
-compr_size(size_t size)
-{
-	size_t ret;
-
-	if (compr_enabled)
-		ret = LZ4_compressBound(size);
-	else
-		ret = size;
-
-	return ret;
-}
-
-static size_t
-compr(uint8_t *in, uint8_t *out, size_t insize, size_t outsize)
-{
-	int ret;
-
-	if (compr_enabled) {
-		ret = LZ4_compress_default((char *)in, (char *)out, insize,
-		                           outsize);
-		if (ret < 0)
-			errx(1, "LZ4_compress_default failed");
-	} else {
-		ret = insize;
-		memcpy(out, in, insize);
-	}
-
-	return ret;
-}
-
-static size_t
-decompr(uint8_t *in, uint8_t *out, size_t insize, size_t outsize)
-{
-	int ret;
-
-	if (compr_enabled) {
-		ret = LZ4_decompress_safe((char *)in, (char *)out, insize,
-		                          outsize);
-		if (ret < 0)
-			errx(1, "LZ4_decompress_safe failed");
-	} else {
-		ret = insize;
-		memcpy(out, in, insize);
-	}
-
-	return ret;
-}
-
 static void
 print_md(FILE *fp, uint8_t *md, size_t size)
 {
@@ -128,7 +77,7 @@ alloc_snap(void)
 
 	snap = calloc(1, sizeof(*snap));
 	if (snap == NULL)
-		err(1, "calloc");
+		err(1, "%s", __func__);
 	return snap;
 }
 
@@ -138,6 +87,26 @@ free_snap(struct snapshot *snap)
 	free(snap);
 }
 
+/*
+ * The snapshot hash is calculated over the
+ * hash of its block descriptors.
+ */
+static void
+hash_snap(struct snapshot *snap, uint8_t *md)
+{
+	blake2b_state ctx;
+	uint64_t i;
+
+	blake2b_init(&ctx, MDSIZE);
+	for (i = 0; i < snap->nr_blk_descs; i++) {
+		struct blk_desc *blk_desc;
+
+		blk_desc = &snap->blk_desc[i];
+		blake2b_update(&ctx, blk_desc->md, sizeof(blk_desc->md));
+	}
+	blake2b_final(&ctx, md, MDSIZE);
+}
+
 static struct snapshot *
 grow_snap(struct snapshot *snap, uint64_t nr_blk_descs)
 {
@@ -153,7 +122,7 @@ grow_snap(struct snapshot *snap, uint64_t nr_blk_descs)
 
 	snap = realloc(snap, size);
 	if (snap == NULL)
-		err(1, "realloc");
+		err(1, "%s", __func__);
 	return snap;
 }
 
@@ -188,7 +157,7 @@ alloc_buf(size_t size)
 
 	p = calloc(1, size);
 	if (p == NULL)
-		err(1, "calloc");
+		err(1, "%s", __func__);
 	return p;
 }
 
@@ -201,19 +170,24 @@ free_buf(uint8_t *buf)
 static void
 hash_blk(uint8_t *buf, size_t size, uint8_t *md)
 {
-	SHA256_CTX ctx;
+	blake2b_state ctx;
 
-	SHA256_Init(&ctx);
-	SHA256_Update(&ctx, buf, size);
-	SHA256_Final(md, &ctx);
+	blake2b_init(&ctx, MDSIZE);
+	blake2b_update(&ctx, buf, size);
+	blake2b_final(&ctx, md, MDSIZE);
 }
 
 static void
 read_blk(uint8_t *buf, struct blk_desc *blk_desc)
 {
+	ssize_t n;
+
 	xlseek(sfd, blk_desc->offset, SEEK_SET);
-	if (xread(sfd, buf, blk_desc->size) == 0)
-		errx(1, "read: unexpected EOF");
+	n = xread(sfd, buf, blk_desc->size);
+	if (n == 0)
+		errx(1, "%s: unexpected EOF", __func__);
+	if (n != blk_desc->size)
+		errx(1, "%s: short read", __func__);
 }
 
 static void
@@ -274,24 +248,20 @@ dedup(int fd, char *msg)
 {
 	struct snapshot *snap;
 	struct chunker *chunker;
-	SHA256_CTX ctx;
-	ssize_t n;
 
 	snap = alloc_snap();
-	chunker = alloc_chunker(fd, BLKSIZE_MAX);
+	chunker = alloc_chunker(fd, BLKSIZE_MIN, BLKSIZE_MAX,
+	                        HASHMASK_BITS, WINSIZE);
 
-	SHA256_Init(&ctx);
-	while ((n = fill_chunker(chunker)) > 0) {
+	while (fill_chunker(chunker) > 0) {
 		uint8_t *chunkp;
 		size_t chunk_size;
 
 		chunkp = get_chunk(chunker, &chunk_size);
-		SHA256_Update(&ctx, chunkp, chunk_size);
 		snap = grow_snap(snap, snap->nr_blk_descs + 1);
 		dedup_chunk(snap, chunkp, chunk_size);
 		drain_chunker(chunker);
 	}
-	SHA256_Final(snap->md, &ctx);
 
 	if (snap->nr_blk_descs > 0) {
 		if (msg != NULL) {
@@ -303,7 +273,7 @@ dedup(int fd, char *msg)
 			memcpy(snap->msg, msg, size);
 			snap->msg[size - 1] = '\0';
 		}
-
+		hash_snap(snap, snap->md);
 		append_snap(snap);
 	}
 
@@ -345,10 +315,8 @@ extract(struct snapshot *snap, void *arg)
 static int
 check_snap(struct snapshot *snap, void *arg)
 {
-	uint8_t md[MDSIZE];
 	uint8_t *buf;
 	int *ret = arg;
-	SHA256_CTX ctx;
 	uint64_t i;
 
 	if (verbose > 0) {
@@ -359,14 +327,16 @@ check_snap(struct snapshot *snap, void *arg)
 
 	buf = alloc_buf(compr_size(BLKSIZE_MAX));
 	for (i = 0; i < snap->nr_blk_descs; i++) {
+		uint8_t md[MDSIZE];
+		blake2b_state ctx;
 		struct blk_desc *blk_desc;
 
 		blk_desc = &snap->blk_desc[i];
 		read_blk(buf, blk_desc);
 
-		SHA256_Init(&ctx);
-		SHA256_Update(&ctx, buf, blk_desc->size);
-		SHA256_Final(md, &ctx);
+		blake2b_init(&ctx, MDSIZE);
+		blake2b_update(&ctx, buf, blk_desc->size);
+		blake2b_final(&ctx, md, MDSIZE);
 
 		if (memcmp(blk_desc->md, md, sizeof(blk_desc->md)) == 0)
 			continue;
@@ -422,6 +392,7 @@ walk_snap(int (*fn)(struct snapshot *, void *), void *arg)
 {
 	uint64_t i;
 
+	xlseek(ifd, SNAP_HDR_SIZE, SEEK_SET);
 	for (i = 0; i < snap_hdr.nr_snapshots; i++) {
 		struct snapshot *snap;
 		int ret;
@@ -633,7 +604,6 @@ main(int argc, char *argv[])
 	if (cflag) {
 		int ret;
 
-		xlseek(ifd, SNAP_HDR_SIZE, SEEK_SET);
 		ret = 0;
 		walk_snap(check_snap, &ret);
 		if (ret != 0)
@@ -644,7 +614,6 @@ main(int argc, char *argv[])
 	}
 
 	if (lflag) {
-		xlseek(ifd, SNAP_HDR_SIZE, SEEK_SET);
 		walk_snap(list, NULL);
 		term();
 		return 0;
@@ -653,7 +622,6 @@ main(int argc, char *argv[])
 	if (id) {
 		struct extract_args args;
 
-		xlseek(ifd, SNAP_HDR_SIZE, SEEK_SET);
 		str2bin(id, md);
 		args.md = md;
 		args.fd = fd;
